@@ -1,6 +1,6 @@
 ;;; consult-denote.el --- Use Consult in tandem with Denote -*- lexical-binding: t -*-
 
-;; Copyright (C) 2024-2025  Free Software Foundation, Inc.
+;; Copyright (C) 2024-2026  Free Software Foundation, Inc.
 
 ;; Author: Protesilaos Stavrou <info@protesilaos.com>
 ;; Maintainer: Protesilaos Stavrou <info@protesilaos.com>
@@ -121,22 +121,25 @@ With optional HAS-IDENTIFIER, only show candidates that have an
 identifier.
 
 Return the absolute path to the matching file."
-  (let* ((single-dir-p (denote-has-single-denote-directory-p))
+  (let* ((roots (denote-directories))
+         (single-dir-p (null (cdr roots)))
          (default-directory (if single-dir-p ; setting the `default-directory' is needed for the preview
-                                (car (denote-directories))
+                                (car roots)
                               (denote-directories-get-common-root)))
+         (files (denote-directory-files
+                 (or denote-file-prompt-use-files-matching-regexp files-matching-regexp)
+                 :omit-current nil nil roots))
          (relative-files (mapcar
-                          #'denote-get-file-name-relative-to-denote-directory
-                          (denote-directory-files
-                           (or denote-file-prompt-use-files-matching-regexp files-matching-regexp)
-                           :omit-current nil nil has-identifier)))
+                          (lambda (file)
+                            (denote-get-file-name-relative-to-denote-directory file))
+                          files))
          (prompt (if single-dir-p
                      (format "%s: " (or prompt-text "Select FILE"))
                    (format "%s in %s: "
                            (or prompt-text "Select FILE")
                            (propertize default-directory 'face 'denote-faces-prompt-current-name))))
          (input (consult--read
-                 (denote--completion-table 'file relative-files)
+                 (apply 'denote-get-completion-table relative-files denote-file-prompt-extra-metadata)
                  :state (consult--file-preview)
                  :require-match (unless no-require-match :require-match)
                  :history 'denote-file-history
@@ -164,53 +167,66 @@ With optional PROMPT-TEXT use it instead of a generic prompt.
 
 With optional FILES-WITH-SEQUENCES as a list of strings, use them as
 completion candidates.  Else use `denote-sequence-get-all-files'."
-  (if-let* ((files (or files-with-sequences (denote-sequence-get-all-files)))
-            (single-dir-p (denote-has-single-denote-directory-p))
-            (relative-files (if single-dir-p
-                                (mapcar #'denote-get-file-name-relative-to-denote-directory files)
-                              files))
-            (prompt (format-prompt (or prompt-text "Select FILE with sequence") nil))
-            (input (consult--read
-                    (denote--completion-table 'file relative-files)
-                    :state (consult--file-preview)
-                    :require-match nil
-                    :history 'denote-sequence-file-history
-                    :prompt prompt)))
-      (if single-dir-p
-          (expand-file-name input (car (denote-directories)))
-        input)
-    (error "There are no sequence notes in the `denote-directory'")))
+  (let* ((roots (denote-directories))
+         (single-dir-p (null (cdr roots)))
+         ;; Some external program may use `default-directory' with the
+         ;; relative file paths of the completion candidates.
+         (default-directory (if single-dir-p
+                                (car roots)
+                              (denote-directories-get-common-root))))
+    (if-let* ((files (or files-with-sequences (denote-sequence-get-all-files)))
+              (relative-files (if single-dir-p
+                                  (mapcar
+                                   (lambda (file)
+                                     (denote-get-file-name-relative-to-denote-directory file))
+                                   files)
+                                files))
+              (prompt (format-prompt (or prompt-text "Select FILE with sequence") nil))
+              (input (consult--read
+                      (apply 'denote-get-completion-table relative-files denote-sequence-file-prompt-extra-metadata)
+                      :state (consult--file-preview)
+                      :require-match nil
+                      :history 'denote-sequence-file-history
+                      :prompt prompt)))
+        (if single-dir-p
+            (expand-file-name input (car roots))
+          input)
+      (error "There are no sequence notes in the `denote-directory'"))))
 
 (defun consult-denote-select-linked-file-prompt (files &optional prompt-text)
   "Prompt for linked file among FILES and use optional PROMPT-TEXT."
-  (let* ((single-dir-p (denote-has-single-denote-directory-p))
-         (file-names (if single-dir-p
-                         (mapcar #'denote-get-file-name-relative-to-denote-directory files)
-                       files))
+  (let* ((roots (denote-directories))
+         (single-dir-p (null (cdr roots)))
+         (relative-files (if single-dir-p
+                             (mapcar
+                              (lambda (file)
+                                (denote-get-file-name-relative-to-denote-directory file))
+                              files)
+                           files))
          (prompt (format-prompt (or prompt-text "Find linked file") nil))
          (input (consult--read
-                 (denote--completion-table 'file file-names)
+                 (apply 'denote-get-completion-table relative-files denote-file-prompt-extra-metadata)
                  :state (consult--file-preview)
                  :require-match t
                  :history 'denote-link-find-file-history
                  :prompt prompt)))
     (if single-dir-p
-        (expand-file-name input (car (denote-directories)))
+        (expand-file-name input (car roots))
       input)))
 
-(defvar denote-silo-extras-directory-history)
-(defvar denote-silo-extras-directories)
+(defvar denote-silo-directory-history)
+(defvar denote-silo-directories)
 
 (defun consult-denote-silo-directory-prompt ()
-  "Like the `denote-silo-extras-directory-prompt' with Consult preview."
-  (let ((default (car denote-silo-extras-directory-history)))
+  "Like the `denote-silo-directory-prompt' with Consult preview."
+  (let ((default (car denote-silo-directory-history)))
     (consult--read
-     (denote--completion-table 'file denote-silo-extras-directories)
+     (denote-get-completion-table denote-silo-directories '(category . file))
      :state (consult--file-preview)
      :require-match t
      :prompt (format-prompt "Select a silo" default)
      :default default
-     :history 'denote-silo-extras-directory-history)))
+     :history 'denote-silo-directory-history)))
 
 (declare-function denote-org--get-outline "denote-org" (file))
 
@@ -222,12 +238,13 @@ completion candidates.  Else use `denote-sequence-get-all-files'."
 FILE has the same meaning as in `denote-org-extras-outline-prompt'."
   (let ((current-file (or file buffer-file-name)))
     (consult--read
-     (denote--completion-table-no-sort 'imenu (denote-org--get-outline current-file))
+     (denote-get-completion-table (denote-org--get-outline current-file) '(category . imenu))
      :state (lambda (_action candidate)
               (with-current-buffer (current-buffer)
-                (when-let* ((_ candidate)
-                            (line (string-to-number (car (split-string candidate)))))
-                  (forward-line (- line 1)))))
+                (save-excursion
+                  (when-let* ((_ candidate)
+                              (line (string-to-number (car (split-string candidate)))))
+                    (forward-line (- line 1))))))
      :prompt (format "Select heading inside `%s': " (propertize (file-name-nondirectory current-file) 'face 'denote-faces-prompt-current-name))
      :require-match t)))
 
@@ -312,15 +329,15 @@ FILE has the same meaning as in `denote-org-extras-outline-prompt'."
 
 (with-eval-after-load 'denote-silo-extras
   (setq consult-denote-silo-source
-    `( :name "Denote silos"
-       :narrow ?L
-       :category file
-       :default t
-       :face consult-denote-directory
-       :history consult-denote-buffer-history
-       :action ,#'dired
-       :state ,#'consult--file-state
-       :items ,denote-silo-extras-directories)))
+        `( :name "Denote silos"
+           :narrow ?L
+           :category file
+           :default t
+           :face consult-denote-directory
+           :history consult-denote-buffer-history
+           :action ,#'dired
+           :state ,#'consult--file-state
+           :items ,denote-silo-extras-directories)))
 
 (declare-function denote-org-outline-prompt "denote-org" (&optional file))
 (declare-function denote-silo-directory-prompt "denote-silo" ())
